@@ -2,23 +2,21 @@
 # this repository contains the full copyright notices and license terms.
 import datetime
 from collections import deque, defaultdict
+from functools import reduce
 from heapq import heappop, heappush
 
-from trytond.model import ModelSQL, fields
+from trytond.model import ModelSQL, fields, tree
 from trytond.wizard import Wizard, StateTransition
 from trytond.pyson import Eval
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 
-__all__ = ['Work', 'PredecessorSuccessor', 'Leveling']
-
 
 def intfloor(x):
     return int(round(x, 4))
 
 
-class Work:
-    __metaclass__ = PoolMeta
+class Work(tree(parent='successors'), metaclass=PoolMeta):
     __name__ = 'project.work'
     predecessors = fields.Many2Many('project.predecessor_successor',
         'successor', 'predecessor', 'Predecessors',
@@ -74,11 +72,6 @@ class Work:
     def __setup__(cls):
         super(Work, cls).__setup__()
 
-    @classmethod
-    def validate(cls, works):
-        super(Work, cls).validate(works)
-        cls.check_recursion(works, parent='successors')
-
     @staticmethod
     def default_leveling_delay():
         return 0.0
@@ -263,12 +256,20 @@ class Work:
 
     def compute_dates(self):
         values = {}
-        get_early_finish = lambda work: values.get(work, {}).get(
-            'early_finish_time', work.early_finish_time)
-        get_late_start = lambda work: values.get(work, {}).get(
-            'late_start_time', work.late_start_time)
-        maxdate = lambda x, y: x and y and max(x, y) or x or y
-        mindate = lambda x, y: x and y and min(x, y) or x or y
+
+        def get_early_finish(work):
+            return values.get(work, {}).get(
+                'early_finish_time', work.early_finish_time)
+
+        def get_late_start(work):
+            return values.get(work, {}).get(
+                'late_start_time', work.late_start_time)
+
+        def maxdate(x, y):
+            return x and y and max(x, y) or x or y
+
+        def mindate(x, y):
+            return x and y and min(x, y) or x or y
 
         # propagate constraint_start_time
         constraint_start = reduce(maxdate, (pred.early_finish_time
@@ -434,7 +435,8 @@ class Work:
         # write values
         write_fields = ('early_start_time', 'early_finish_time',
                         'late_start_time', 'late_finish_time')
-        for work, val in values.iteritems():
+        to_write = []
+        for work, val in values.items():
             write_cond = False
             for field in write_fields:
                 if field in val and getattr(work, field) != val[field]:
@@ -442,15 +444,20 @@ class Work:
                     break
 
             if write_cond:
-                self.write([work], val)
+                to_write.extend(([work], val))
+        if to_write:
+            self.write(*to_write)
 
     def reset_leveling(self):
-        get_key = lambda w: (set(p.id for p in w.predecessors),
-                             set(s.id for s in w.successors))
+        def get_key(w):
+            return (
+                set(p.id for p in w.predecessors),
+                set(s.id for s in w.successors))
 
-        parent_id = self.parent and self.parent.id or None
+        if not self.parent:
+            return
         siblings = self.search([
-                ('parent', '=', parent_id)
+                ('parent', '=', self.parent.id)
                 ])
         to_clean = []
 
@@ -469,14 +476,18 @@ class Work:
 
     def create_leveling(self):
         # define some helper functions
-        get_key = lambda w: (set(p.id for p in w.predecessors),
-                             set(s.id for s in w.successors))
-        over_alloc = lambda current_alloc, work: (
-            reduce(lambda res, alloc: (res
+        def get_key(w):
+            return (
+                set(p.id for p in w.predecessors),
+                set(s.id for s in w.successors))
+
+        def over_alloc(current_alloc, work):
+            return reduce(lambda res, alloc: (
+                    res
                     or (current_alloc[alloc.employee.id]
                         + alloc.percentage) > 100),
                 work.allocations,
-                False))
+                False)
 
         def sum_allocs(current_alloc, work):
             res = defaultdict(float)
@@ -531,15 +542,12 @@ class Work:
         siblings = [s for s in siblings if get_key(s) == refkey]
 
         for sibling, delay in compute_delays(siblings):
-            self.write([sibling], {
-                    'leveling_delay': delay,
-                    })
+            sibling.leveling_delay = delay
 
         siblings.reverse()
         for sibling, delay in compute_delays(siblings):
-            self.write([sibling], {
-                    'back_leveling_delay': delay,
-                    })
+            sibling.back_leveling_delay = delay
+        self.__class__.save(siblings)
 
         if self.parent:
             self.parent.compute_dates()
@@ -596,7 +604,7 @@ class PredecessorSuccessor(ModelSQL):
         super(PredecessorSuccessor, cls).write(*args)
 
         work_ids = [v for values in args[1::2]
-            for k, v in values.itervalues()
+            for k, v in values.values()
             if k in ('predecessor', 'successor')]
         works = Work.browse(work_ids)
         for work in works:
